#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <windows.h>
#include <commdlg.h>

static uint8_t firmware[12288];
static int enable;

/*
 * The config dialog loop (open md_firmware_res.res in a resource editor)
 */
typedef enum { FW_F8AUTO, FW_F8ON, FW_F8OFF } FW_F8;
char *fw_bool[2]={"Off", "On"};
char *fw_mtrs[3]={"Auto", "On", "Off"};
char fw_rom[260];
BOOL fw_disengage;
FW_F8 fw_f8;

BOOL WINAPI DialogProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
 int x;
 
 FILE *file;
 OPENFILENAME ofn;
 char browsefn[260];
 char erstr[1024];

 char *foo_p;
 long  foo_l;
 HWND  foo_c;
 
 /* wParam = what was hit */
 switch(uMsg)
 {
  case WM_INITDIALOG:
   if (fw_disengage) CheckDlgButton(hWindow, 1, BST_CHECKED);
   CheckDlgButton(hWindow, fw_f8+3, BST_CHECKED);
   SetWindowText(GetDlgItem(hWindow, 9), fw_rom);
   return TRUE;
  case WM_COMMAND:
   switch (wParam)
   {
    case 2: /* ESC */
    case 7: /* Cancel button */
     EndDialog(hWindow, 0);
     DestroyWindow(hWindow);
     return TRUE;
    case 6: /* OK clicked */
     fw_disengage=IsDlgButtonChecked(hWindow, 1);
     if (IsDlgButtonChecked(hWindow, 3)) fw_f8=FW_F8AUTO;
     if (IsDlgButtonChecked(hWindow, 4)) fw_f8=FW_F8ON;
     if (IsDlgButtonChecked(hWindow, 5)) fw_f8=FW_F8OFF;
     
     foo_c=GetDlgItem(hWindow, 9);
     foo_l=GetWindowTextLength(foo_c);
     foo_p=malloc(foo_l+1);
     GetWindowText(foo_c,foo_p,foo_l+1);
     strcpy(browsefn,foo_p);
     free(foo_p);
     
     file=fopen(browsefn,"rb");
     if (file)
     {
      long l;
      
      fseek(file,0,SEEK_END);
      l=ftell(file);
      fclose(file);
      if (l!=12288)
      {
       sprintf(erstr, "The size of the file '%s' does not match the ROM capacity.", browsefn);
       MessageBox(hWindow,erstr,"Firmware Card",48);
       return TRUE;
      }
     }
     else
     {
      sprintf(erstr, "The file '%s' could not be opened:\n\n%s ($%04X)",
              browsefn, strerror(errno), errno);
      MessageBox(hWindow,erstr,"Firmware Card",48);
      return TRUE;
     }
     
     strcpy(fw_rom, browsefn);
     EndDialog(hWindow, 0);
     DestroyWindow(hWindow);
     return TRUE;
    case 10: /* Browse clicked */
     *browsefn=0;
     ZeroMemory(&ofn,sizeof(ofn));
     ofn.lStructSize    =sizeof(ofn);
     ofn.hwndOwner      =hWindow;
     ofn.lpstrFile      =browsefn;
     ofn.nMaxFile       =260;
     ofn.lpstrFilter    ="ROM Files (*.ROM, *.BIN)\0*.ROM,*.BIN\0All Files (*.*)\0*.*\0";
     ofn.nFilterIndex   =1;
     ofn.lpstrFileTitle =0;
     ofn.nMaxFileTitle  =0;
     ofn.lpstrInitialDir=0;
     ofn.Flags          =(OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST);
     if (GetOpenFileName(&ofn))
     {
      foo_c=GetDlgItem(hWindow, 9);
      SetWindowText(foo_c,browsefn);
     }
     return TRUE;
   }
   break;
  case WM_CLOSE:
   EndDialog(hWindow, 0);
   DestroyWindow(hWindow);
   return TRUE;
 }

 return FALSE;
}

/*
 * Module information
 */
__declspec(dllexport) uint32_t mod_type=0x544F4C53;
__declspec(dllexport) uint8_t *mod_name="Integer Basic Firmware Card";
__declspec(dllexport) uint8_t *mod_pref="firmware";
__declspec(dllexport) uint8_t *mod_vers="0.02";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas\n";

__declspec(dllexport) uint8_t  *firmware_preload_name=0;
__declspec(dllexport) uint16_t firmware_preload_size=0;
__declspec(dllexport) uint8_t  *firmware_preload_ptr=0;

/*
 * The firmware card is activated when an EVEN address in slot i/o space is
 * WRITTEN to, and deactivated when an ODD address is WRITTEN to.
 */
__declspec(dllexport) void firmware_write16 (uint16_t a, uint8_t v)
{
 enable=1-(a&1);
}

/*
 * We are a slot 0 card; we don't have a rom space.
 */
__declspec(dllexport) void firmware_write256 (uint16_t a, uint8_t v)
{
}

/*
 * We don't have a shared memory space.
 */
__declspec(dllexport) void firmware_write2048 (uint16_t a, uint8_t v)
{
}

/*
 * Reading I/O space has no effect.
 */
__declspec(dllexport) uint8_t firmware_read16 (uint16_t a)
{
 return 255;
}

/*
 * We are a slot 0 card; we don't have a rom space.
 */
__declspec(dllexport) uint8_t firmware_read256 (uint16_t a)
{
 return 255;
}

/*
 * We don't have a shared memory space.
 */
__declspec(dllexport) uint8_t firmware_read2048 (uint16_t a)
{
 return 255;
}

/*
 * No change to what writing to any memory address will do.
 */
__declspec(dllexport) int16_t firmware_override_wr (uint16_t a, uint8_t v)
{
 return -1;
}

/*
 * Intercept READS to the top 12K of memory, IF enabled.
 * Otherwise do nothing.
 */
__declspec(dllexport) int16_t firmware_override_rd (uint16_t a)
{
 if (a>=0xD000)
 {
  if (a>=0xF800)
  {
   if (fw_f8==FW_F8ON) return firmware[a-0xD000];
   if (fw_f8==FW_F8OFF) return -1;
  }
  if (!enable) return -1;
  return firmware[a-0xD000];
 }
 return -1;
}

/*
 * Disable ourself when Reset is hit.
 */
__declspec(dllexport) int firmware_reset (void)
{
 if (fw_disengage) enable=0;
 return 0;
}

static char inifile[1024];

/*
 * Initialize the driver.
 */
__declspec(dllexport) int firmware_init (void)
{
 FILE *file;
 int l;
 char modbuf[256];

 /* Where is our INI file? */
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 
 l=GetPrivateProfileString("mod_firmware","DisengageOnReset","On",modbuf,256,inifile);
 fw_disengage=(!strcasecmp(modbuf,"On"))?TRUE:FALSE;
 
 l=GetPrivateProfileString("mod_firmware","MonitorRom","Auto",modbuf,256,inifile);
 if (!strcasecmp(modbuf,"Auto")) fw_f8=FW_F8AUTO;
 if (!strcasecmp(modbuf,"On")) fw_f8=FW_F8ON;
 if (!strcasecmp(modbuf,"Off")) fw_f8=FW_F8OFF;
 
 l=GetPrivateProfileString("mod_firmware","RomFilename","intbasic.rom",fw_rom,260,inifile);

 file=fopen(fw_rom,"rb");
 if (!file)
 {
  perror(fw_rom);
  return -1;
 }
 fread(firmware,1,12288,file);
 fclose(file);
 printf("mod_firmware: Initialized with firmware file '%s'\n",fw_rom);
 
 enable=0;
 return 0;
}

/*
 * Nothing to do when deinitializing.
 */
__declspec(dllexport) int firmware_deinit (void)
{
 int l;
 
 l=WritePrivateProfileString("mod_firmware","DisengageOnReset",fw_bool[fw_disengage],inifile);
 l=WritePrivateProfileString("mod_firmware","MonitorRom",fw_mtrs[fw_f8],inifile);
 l=WritePrivateProfileString("mod_firmware","RomFilename",fw_rom,inifile);
 
 return 0;
}

static HINSTANCE hinstance;

__declspec(dllexport) void firmware_config (void)
{
 FILE *file;
 int e;
 
 e=DialogBox(hinstance, "DLGFWCARD", 0, DialogProc);
 
 file=fopen(fw_rom,"rb");
 if (!file)
 {
  MessageBox(0,"FATAL: 'RomFilename' data inconsistency", "Firmware Card", 16);
  exit(1);
 }
 fread(firmware,1,12288,file);
 fclose(file);
}

/*
 * Needed for DialogBox to work
 */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved)
{
 hinstance=hinstDLL;
 return TRUE;
}
